
#define	NEKO

#include	<windows.h>
#include	"fddfile.h"
#include	"fdd_d88.h"

#ifdef KEROPI
#include	"../win32/common.h"
#include	"../win32/fileio.h"
#include	"fdc.h"

#define		FILEHDL		FILEH
#define		FILE_OPEN	File_Open
#define		FILE_READ	File_Read
#define		FILE_SEEK	File_Seek
#define		FILE_CLOSE	File_Close
#endif

#ifdef NEKO
#include	"common.h"
#include	"np2.h"
#include	"dosio.h"
#include	"iobridge.h"
#include	"fdc.h"
#include	"trace.h"

#define		FILEHDL		FILEH
#define		FILE_OPEN	file_open
#define		FILE_READ	file_lread
#define		FILE_WRITE	file_lwrite
#define		FILE_SEEK	file_seek
#define		FILE_CLOSE	file_close
#endif

#define		D88BUFSIZE		0x4000
#define		D88TRACKMAX		10600

static	D88_HEADER	d88head[4];
static	D88_SECTOR	cursec;
static	BYTE		curdrv = 0xff;
static	WORD		curtrk = 0xffff;
static	BYTE		curwrite = 0;
static	DWORD		curfp = 0;
static	DWORD		cursize = 0;
static	BYTE		curtrkaccess = 0;
static	BYTE		curtype = 0;

static	BYTE		D88_BUF[D88BUFSIZE];

// ------------------------------------------------------------------------

static BOOL curdataflush(void) {

	FILEHDL	hdl;
	BOOL	ret = SUCCESS;

	if ((!curfp) || (!cursize) || (!curwrite)) {
		return(SUCCESS);
	}
	curwrite = 0;
	if ((hdl = FILE_OPEN(fddfile[curdrv].fname)) == (FILEHDL)-1) {
		return(FAILURE);
	}
	if ((FILE_SEEK(hdl, curfp, FSEEK_SET) != curfp) ||
		(FILE_WRITE(hdl, D88_BUF, cursize) != cursize)) {
		ret = FAILURE;
	}
	if (FILE_CLOSE(hdl)) {
		ret = FAILURE;
	}
	return(ret);
}


static void drvflush(BYTE drv) {

	if (curdrv == drv) {
		curdataflush();
		curdrv = 0xff;
		curtrk = 0xffff;
	}
}

static DWORD nexttrackp(D88_HEADER *head, DWORD fptr, DWORD last) {

	int		t;
	DWORD	ret;
	DWORD	*trkp;

	ret = last;
	trkp = (DWORD *)head->trackp;
	for (t=0; t<164; t++, trkp++) {
		if ((*trkp > fptr) && (*trkp < ret)) {
			ret = *trkp;
		}
	}
	return(ret);
}

static BOOL read_d88track(BYTE drv, BYTE track, BYTE type) {

	FILEHDL	hdl;

	curdataflush();
	curdrv = drv;
	curtrk = track;
	curtype = type;
	curwrite = 0;
	curtrkaccess = SUCCESS;

	if ((drv >= 4) || (track > 163) ||
		(type != (d88head[drv].fd_type >> 4)) ||
		((curfp = d88head[drv].trackp[track]) == 0) ||
		((cursize = nexttrackp(&d88head[drv], curfp,
							d88head[drv].fd_size) - curfp) > D88BUFSIZE)) {
		goto readd88_err;
	}
	if ((hdl = FILE_OPEN(fddfile[drv].fname)) == (FILEH)-1) {
		goto readd88_err;
	}
	if (FILE_SEEK(hdl, curfp, FSEEK_SET) != curfp) {
		FILE_CLOSE(hdl);
		goto readd88_err;
	}
	if (FILE_READ(hdl, D88_BUF, cursize) != cursize) {
		FILE_CLOSE(hdl);
		goto readd88_err;
	}
	if (FILE_CLOSE(hdl)) {
		goto readd88_err;
	}
	return(SUCCESS);

readd88_err:
	curfp = 0;
	cursize = 0;
	curtrkaccess = FAILURE;
	return(FAILURE);
}

static BOOL seek_d88(BYTE drv, BYTE track) {

	if ((curdrv != drv) || (curtrk != track) || (curtype != CTRL_FDMEDIA)) {
		read_d88track(drv, track, CTRL_FDMEDIA);
	}
	return(curtrkaccess);
}

static D88_SECTOR *searchsector_d88(void) {

	BYTE	*p;
	BYTE	sec;
	WORD	pos = 0;

	p = D88_BUF;
	for (sec=0; sec<40; ) {
		if ((((D88_SECTOR *)p)->c == fdc.C) &&
			(((D88_SECTOR *)p)->h == fdc.H) &&
			(((D88_SECTOR *)p)->r == fdc.R) &&
			(((D88_SECTOR *)p)->n == fdc.N)) {
			return((D88_SECTOR *)p);
		}
		if (++sec >= ((D88_SECTOR *)p)->sectors) {
			break;
		}

		pos += ((D88_SECTOR *)p)->size;
		pos += sizeof(D88_SECTOR);

		p += ((D88_SECTOR *)p)->size;
		p += sizeof(D88_SECTOR);
	}
	return(NULL);
}

// ------------------------------------------------------------------------

BOOL fdd_set_d88(BYTE drv, LPSTR fname) {

	WORD	attr;
	FILEHDL	hdl;
	DWORD	rsize;

	fdd_eject_d88(drv);
	if ((attr = file_attr(fname)) & 0x18) {
		return(FAILURE);
	}
	if ((hdl = FILE_OPEN(fname)) == (FILEH)-1) {
		return(FAILURE);
	}
	rsize = FILE_READ(hdl, &d88head[drv], sizeof(D88_HEADER));
	FILE_CLOSE(hdl);
	if (rsize != sizeof(D88_HEADER)) {
		fdd_eject_d88(drv);
		return(FAILURE);
	}
	strcpy(fddfile[drv].fname, fname);
	fddfile[drv].type = DISKTYPE_D88;
	fddfile[drv].protect = FALSE;
	if ((attr & 1) || (d88head[drv].protect & 0x10)) {
		fddfile[drv].protect = TRUE;
	}
	return(SUCCESS);
}

BOOL fdd_eject_d88(BYTE drv) {

	drvflush(drv);
	ZeroMemory(&d88head[drv], sizeof(D88_HEADER));
	fddfile[drv].fname[0] = '\0';
	fddfile[drv].type = DISKTYPE_NOTREADY;
	return(SUCCESS);
}


BOOL fdd_diskaccess_d88(void) {

	if (CTRL_FDMEDIA != (d88head[fdc.us].fd_type >> 4)) {
		return(FAILURE);
	}
	return(SUCCESS);
}

BOOL fdd_seek_d88(void) {

	if (fdc.ncn >= 82) {
		return(FAILURE);
	}
	return(seek_d88(fdc.us, (fdc.ncn << 1) + fdc.hd));
}

BOOL fdd_seeksector_d88(void) {

	if (fdc.treg[fdc.us] >= 82) {
		return(FAILURE);
	}
	if (seek_d88(fdc.us, (fdc.treg[fdc.us] << 1) + fdc.hd)) {
		return(FAILURE);
	}
	if (!searchsector_d88()) {
		return(FAILURE);
	}
	return(SUCCESS);
}

BOOL fdd_read_d88(void) {

	D88_SECTOR	*p;
	WORD		size;

	if (fdc.treg[fdc.us] >= 82) {
		return(FAILURE);
	}
	if (seek_d88(fdc.us, (fdc.treg[fdc.us] << 1) + fdc.hd)) {
		return(FAILURE);
	}
	p = searchsector_d88();
	if (!p) {
		return(FAILURE);
	}
	if (fdc.N < 8) {
		size = 128 << fdc.N;
	}
	else {
		size = 128 << 8;
	}
	fdc.bufcnt = size;
	ZeroMemory(fdc.buf, size);
	if (size > p->size) {
		size = p->size;
	}
	if (size) {
		memcpy(fdc.buf, p+1, size);
	}
	return(SUCCESS);
}

BOOL fdd_write_d88(void) {

	D88_SECTOR	*p;
	WORD		size;

	if (fdc.treg[fdc.us] >= 82) {
		return(FAILURE);
	}
	if (seek_d88(fdc.us, (fdc.treg[fdc.us] << 1) + fdc.hd)) {
		return(FAILURE);
	}
	p = searchsector_d88();
	if (!p) {
		return(FAILURE);
	}
	if (fdc.N < 8) {
		size = 128 << fdc.N;
	}
	else {
		size = 128 << 8;
	}
	if (size > p->size) {
		size = p->size;
	}
	if (size) {
		memcpy(p+1, fdc.buf, size);
		curwrite = TRUE;
	}
	return(SUCCESS);
}

BOOL fdd_readid_d88(void) {

	BYTE	*p;
	BYTE	sec;

	if (fdc.treg[fdc.us] >= 82) {
		return(FAILURE);
	}
	if (seek_d88(fdc.us, (fdc.treg[fdc.us] << 1) + fdc.hd)) {
		return(FAILURE);
	}
	p = D88_BUF;
	for (sec=0; sec<40; ) {
		if (sec == fdc.crcn) {
			fdc.C = ((D88_SECTOR *)p)->c;
			fdc.H = ((D88_SECTOR *)p)->h;
			fdc.R = ((D88_SECTOR *)p)->r;
			fdc.N = ((D88_SECTOR *)p)->n;
			fdc.crcn++;
#if 1
			if (fdc.crcn >= ((D88_SECTOR *)p)->sectors) {
				fdc.crcn = 0;
			}
#endif
			return(SUCCESS);
		}
		if (++sec >= ((D88_SECTOR *)p)->sectors) {
			break;
		}
		p += ((D88_SECTOR *)p)->size;
		p += sizeof(D88_SECTOR);
	}
	fdc.crcn = 0xff;
	return(FAILURE);
}

// --------------------------------------------------------------------------

static BOOL formating = FALSE;
static BYTE formatsec = 0;
static BYTE formatwrt = 0;
static WORD formatpos = 0;

static int fileappend(FILEH hdl, D88_HEADER *head,
									long ptr, long last, long apsize) {

	long	length;
	WORD	size;
	WORD	rsize;
	int		t;
	long	*trkp;
	BYTE	tmp[0x4000];

	if ((length = last - ptr) <= 0) {			// KvȂ
		return(0);
	}
	while(length) {
		if (length >= sizeof(tmp)) {
			size = sizeof(tmp);
		}
		else {
			size = (WORD)length;
		}
		length -= (long)size;
		file_seek(hdl, ptr+length, 0);
		rsize = file_read(hdl, tmp, size);
		file_seek(hdl, ptr+length+apsize, 0);
		file_write(hdl, tmp, rsize);
	}

	trkp = head->trackp;
	for (t=0; t<164; t++, trkp++) {
		if ((*trkp) && (*trkp >= ptr)) {
			(*trkp) += apsize;
		}
	}
	return(0);
}


static void endoftrack(WORD fmtsize, BYTE sectors) {

	FILEH		hdl;
	int			i;
	WORD		trk;
	D88_HEADER	*head;
	long		fpointer;
	long		endpointer;
	long		lastpointer;
	long		trksize;
	int			ptr;
	long		apsize;

	head = &d88head[fdc.us];
	trk = (fdc.treg[fdc.us] << 1) + fdc.hd;

	ptr = 0;
	for (i=0; i<(int)sectors; i++) {
		((D88_SECTOR *)(&D88_BUF[ptr]))->sectors = sectors;
		ptr += ((D88_SECTOR *)(&D88_BUF[ptr]))->size + sizeof(D88_SECTOR);
	}

	if ((hdl = FILE_OPEN(fddfile[fdc.us].fname)) == (FILEH)-1) {
		return;
	}
	lastpointer = file_seek(hdl, 0, 2);
	if ((fpointer = head->trackp[trk]) == 0) {
		fpointer = 0;								// VKgbN
		for (i=trk; i>=0; i--) {
			if (head->trackp[i]) {
				fpointer = head->trackp[i];
				break;
			}
		}
		if (fpointer) {								// qbg
			fpointer = nexttrackp(head, fpointer, lastpointer);
		}
		else {
			fpointer = sizeof(D88_HEADER);
		}
		endpointer = fpointer;
	}
	else {										// gbNf[^͊ɂ
		endpointer = nexttrackp(head, fpointer, lastpointer);
	}
	trksize = endpointer - fpointer;
	if ((apsize = (long)fmtsize - trksize) > 0) {
								// ރf[^̂ف[傫
		fileappend(hdl, head, endpointer, lastpointer, apsize);
		head->fd_size += apsize;
	}
	head->trackp[trk] = fpointer;
	file_seek(hdl, fpointer, 0);
	file_write(hdl, D88_BUF, fmtsize);
	file_seek(hdl, 0, 0);
	file_write(hdl, head, sizeof(D88_HEADER));
	file_close(hdl);
}


BOOL fdd_formatinit_d88(void) {

	if (fdc.treg[fdc.us] < 82) {
		formating = TRUE;
		formatsec = 0;
		formatpos = 0;
		formatwrt = 0;
		curdataflush();
		curdrv = 0xff;
		return(SUCCESS);
	}
	return(FAILURE);
}

	// todo AtH[}bgƂ fBXNP̐؂̂ĂƂc
BOOL fdd_formating_d88(DWORD ID) {

	WORD	size;

	if (!formating) {
		return(FAILURE);
	}
	if (fdc.N < 8) {
		size = 128 << fdc.N;
	}
	else {
		size = 128 << 8;
	}
	if ((formatpos + sizeof(D88_SECTOR) + size) < D88TRACKMAX) {
		ZeroMemory(D88_BUF + formatpos, sizeof(D88_SECTOR));
		*(DWORD *)(&D88_BUF[formatpos]) = ID;
		((D88_SECTOR *)(&D88_BUF[formatpos]))->size = size;
		formatpos += sizeof(D88_SECTOR);
		memset(D88_BUF + formatpos, fdc.d, size);
		formatpos += size;
		formatwrt++;
	}
	formatsec++;
	TRACE_("format sec", formatsec);
	TRACE_("format wrt", formatwrt);
	TRACE_("format max", fdc.sc);
	TRACE_("format pos", formatpos);
	if (formatsec >= fdc.sc) {
		endoftrack(formatpos, formatwrt);
		formating = FALSE;
	}
	return(SUCCESS);
}

BOOL fdd_isformating_d88(void) {

	return(formating);
}

