
typedef struct {
	I286_OUT (*fm_w)(WORD port, BYTE val);
	I286_INB (*fm_r)(WORD port);
	I286_OUT (*pcm_w)(WORD port, BYTE val);
	I286_INB (*pcm_r)(WORD port);
	void (__fastcall *outbuffer)(void *buf, int length);
	void (*extendchannel)(BYTE enable);
} SOUNDBOARDS;

// -------------------------------------------------------------------- system

typedef struct {
	BYTE	opnreg;
	BYTE	extreg;
	BYTE	lowreg;
	BYTE	adpcmmask;
	BYTE	channels;
	BYTE	extend;
	BYTE	reg[0x300];
} OPN_T;

typedef struct {
	WORD	timera;
	BYTE	timerb;
	BYTE	status;
	BYTE	reg;
} FMTIMER_T;

// ----------------------------------------------------------------- opn, opna

#define	OPNA_CLOCK		3579545L

#define	MAXCHANNEL		9

#define	SIN_BITS		10
#define	EVC_BITS		10
#define	ENV_BITS		16
#define	KF_BITS			6
#define	FREQ_BITS		16

#define	ENVTBL_BIT		14
#define	SINTBL_BIT		15

#define	FREQ_RATE		(1L << (21-FREQ_BITS))

#define	TL_BITS			(FREQ_BITS+2)

#define	OPM_OUTSB		(TL_BITS + 2 - 16)			// OPM output 16bit
#define	OPM_OUTSB_8		(OPM_OUTSB + 8)				// OPM output 8bit
#define	OPM_MAXOUT		(0x7fffL<<OPM_OUTSB)
#define	OPM_MINOUT		(-0x8000L<<OPM_OUTSB)

#define	SIN_ENT			(1L << SIN_BITS)
#define	EVC_ENT			(1L << EVC_BITS)

#define	EC_ATTACK		0								// ATTACK start
#define	EC_DECAY		(EVC_ENT << ENV_BITS)			// DECAY start
#define	EC_OFF			((2L * EVC_ENT) << ENV_BITS)	// OFF

#define	EG_STEP			(96.0/EVC_ENT)					// dB step
#define	TL_MAX			(EVC_ENT * 2)

#define	SLOT1			0				// slot number
#define	SLOT2			2
#define	SLOT3			1
#define	SLOT4			3

#define EM_ATTACK		4
#define EM_DECAY1		3
#define EM_DECAY2		2
#define EM_RELEASE		1
#define EM_OFF			0


typedef struct {
	long	*detune1;			// detune1
	WORD	detune2;			// detune2
	WORD	totallevel;			// total level
	long	decaylevel;			// decay level
	long	*attack;			// attack raito
	long	*decay1;			// decay1 raito
	long	*decay2;			// decay2 raito
	long	*release;			// release raito
	long 	freq_cnt;			// frequency count
	long	freq_inc;			// frequency step
	BYTE	keyscale;			// key scale
	BYTE	multiple;			// multiple

	BYTE	env_mode;			// envelope mode
	BYTE	envraito;			// envelope raito
	long	env_cnt;			// envelope count
	long	env_end;			// envelope end count
	long	env_inc;			// envelope step
	long	env_inc_attack;		// envelope attack step
	long	env_inc_decay1;		// envelope decay1 step
	long	env_inc_decay2;		// envelope decay2 step
	long	env_inc_release;	// envelope release step
} FM_SLOT;

typedef struct {
	FM_SLOT		SLOT[4];
	BYTE		algorhythm;			// algorythm
	BYTE		feedback;			// self feedback
	long		op1fb;				// operator1 feedback
	long		*connect1;			// operator1 connect
	long		*connect3;			// operator3 connect
	long		*connect2;			// operator2 connect
	long		*connect4;			// operator4 connect
	WORD		keynote[4];			// key note
	BYTE		keyfunc[4];			// key function
	BYTE		kcode[4];			// key code
	BYTE		pan;				// pan
	BYTE		extop;				// extendopelator-enable
	BYTE		stereo;				// stereo-enable
	BYTE		padding;
} FM_CH;

// ----------------------------------------------------------------------- psg

typedef struct {
	WORD	psgtone_freq;
	WORD	psgtone_cnt;
	BYTE	*psgtone_volp;
	WORD	psgtone_puchi;
} PSGTONE_T;

typedef struct {
	WORD	psgnoise_rand0;
	WORD	psgnoise_rand1;
	WORD	psgnoise_cnt;
	WORD	psgnoise_freq;
	WORD	psgnoise_table[16];
} PSGNOISE_T;

typedef struct {
	WORD	tune1;		// 0
	WORD	tune2;		// 2
	WORD	tune3;		// 4
	BYTE	noise;		// 6
	BYTE	mixer;		// 7
	BYTE	vol1;		// 8
	BYTE	vol2;		// 9
	BYTE	vol3;		// a
	WORD	envtime;	// b
	BYTE	env;		// d
	BYTE	io1;
	BYTE	io2;
} PSGREG_T;

typedef struct {
		PSGTONE_T	psgtone[3];
		PSGNOISE_T	psgnoise;
		PSGREG_T	reg;
		WORD		psgenv_cnt;
		WORD		psgenv_max;
		BYTE		psgenv_mod;
		BYTE		psgenv_vol;
		BYTE		psgenv_volcnt;
		BYTE		psg_mixer;
} PSG_T;

#define		PSGENV_INC			15
#define		PSGENV_ONESHOT		16
#define		PSGENV_LASTON		32
#define		PSGENV_ONECYCLE		64

// -------------------------------------------------------------------- rhythm

#define	RHYTHM_MAX	6

#define C4_DWORD(a,b,c,d) (((DWORD)d<<24) | ((DWORD)c << 16) | ((WORD)b << 8) | (a))

typedef struct {
	DWORD	head;
	DWORD	size;
} WAVE_HEADER;

typedef struct {
	WORD	format;
    WORD	channel;
    DWORD	rate;
    DWORD	rps;
	WORD	block;
	WORD	bit;
} WAVE_INFOS;

typedef struct {
	HANDLE	handle;
	BYTE	*rhythm_data;
	DWORD	rhythm_samples;
	DWORD	rhythm_volume;
	DWORD	rhythm_lr;
	BYTE	*rhythm_ptr;
	DWORD	rhythm_remain;
} RHYTHMFILE_TABLE;

typedef struct {
	RHYTHMFILE_TABLE	r[RHYTHM_MAX];
	WORD				bitmap;
} RHYTHM_TABLE;

// --------------------------------------------------------------------- 86PCM

typedef struct {
		BYTE	buffer[0x8000];
		int		write;
		WORD	readpos;
		WORD	bufsize;
		WORD	fifosize;
		BYTE	extfunc;
		BYTE	lrclock;
		BYTE	fifo;
		BYTE	dactrl;
		BYTE	step;
		BYTE	count;
} PCM86_TABLE;

// -------------------------------------------------------------------- CS4231

typedef struct {
	BYTE	adc_l;				// 0
	BYTE	adc_r;				// 1
	BYTE	aux1_l;				// 2
	BYTE	aux1_r;				// 3
	BYTE	aux2_l;				// 4
	BYTE	aux2_r;				// 5
	BYTE	dac_l;				// 6
	BYTE	dac_r;				// 7
	BYTE	datafmt;			// 8
	BYTE	intconfig;			// 9
	BYTE	pinctrl;			// a
	BYTE	errorstatus;
	BYTE	mode_id;
	BYTE	loopctrl;
	WORD	playcount;

	WORD	featurefunc;
	BYTE	line_l;
	BYTE	line_r;
	WORD	timer;
	BYTE	reserved1;
	BYTE	reserved2;
	BYTE	featurestatus;
	BYTE	chipid;
	BYTE	monoinput;
	BYTE	reserved3;
	BYTE	cap_datafmt;
	BYTE	reserved4;
	WORD	cap_basecount;
} CS4231_REG;

typedef struct {
	BYTE		buffer[256];
	DWORD		bufsize;
	DWORD		readpos;
	DWORD		writepos;
	DWORD		curtime;
	DWORD		step;
	DWORD		pcmdata;
	BYTE		(__fastcall * proc)(void);

	CS4231_REG	reg;
	BYTE		enable;
	BYTE		portctrl;
	WORD		port;
	BYTE		dmairq;
	BYTE		dmach;
	BYTE		adrs;
	BYTE		index;
	BYTE		intflag;
	BYTE		outenable;
} CS4231_TABLE;

// --------------------------------------------------------------------------

extern FMTIMER_T fmtimer;
extern OPN_T opn;
extern SOUNDBOARDS soundboard;
extern SOUNDBOARDS BOARDNONE;
extern SOUNDBOARDS BOARD14;
extern SOUNDBOARDS BOARD26K;
extern SOUNDBOARDS BOARD86;
extern SOUNDBOARDS BOARDx2;
extern SOUNDBOARDS BOARD118;

void soundboard_init(void);
void soundboard_term(void);
void soundboard_reset(BYTE num);
void FMBOARD_setregister(BYTE reg, BYTE value);
I286_OUT sound_dummyout(WORD port, BYTE value);
I286_INB sound_dummyin(WORD port);
void __fastcall makebeeponly(void *buf, int length);



extern	FM_CH	FMCH[MAXCHANNEL];

void FM_inittable(void);
void FM_reset(void);
void FM_setregister(BYTE reg, BYTE value, BYTE ch);
void FM_keyon(int ch, BYTE value);
void FM_setalgorithm(FM_CH *CH);
void __fastcall FM_update(void *buf, int length);



extern	WORD	psg_volume[16];
extern	WORD	psg_puchidec;
extern	DWORD	psg_base;
extern	PSG_T 	psg_1;
extern	PSG_T	psg_2;

void PSG_inittable(void);
void PSG_reset(PSG_T *psg);
void PSG_setregister(PSG_T *psg, BYTE reg);
BYTE PSG_getjoyport(PSG_T *psg);
void __fastcall PSG_update(void *buf, PSG_T *psg, int buflength);



void rhythm_init(void);
void rhythm_term(void);
void rhythm_reset(void);
void rhythm_setregister(BYTE reg, BYTE value);
void __fastcall rhythm_update(void *buf, int buflength);



extern PCM86_TABLE pcm86;

void pcm86init(void);
I286_OUT pcm86_w(WORD port, BYTE val);
I286_INB pcm86_r(WORD port);
void __fastcall PCM86_update(void *buf, int buflength);



extern CS4231_TABLE cs4231;

void cs4231_disable(void);
void cs4231_reset(void);
I286_OUT cs4231p_w(WORD port, BYTE value);
I286_INB cs4231p_r(WORD port);
I286_OUT cs4231_w(WORD port, BYTE value);
I286_INB cs4231_r(WORD port);
BYTE __fastcall cs4231dmafunc(BYTE func);
void __fastcall cs4231_update(void *buf, int buflength);



void opn_init(void);
void opna_init(void);
void opnx2_init(void);
void ymf_init(void);

