#include	<windows.h>
#include	"common.h"
#include	"dsound3.h"
#include	"iobridge.h"
#include	"fmboard.h"
#include	"dosio.h"


RHYTHM_TABLE		rhythm;

static int waveread(RHYTHMFILE_TABLE *item, LPSTR filename, DWORD samplerate) {

	FILEH			src;
	WAVE_HEADER		whead;
	WAVE_INFOS		winfo;
	long			seekp;
	int				step;
	HANDLE			hwave;
	BYTE			*wave;
	DWORD			i;

	if ((src = file_open_c(filename)) == (FILEH)-1) {
		return(FAILURE);
	}
	if ((file_read(src, &whead, sizeof(whead)) != sizeof(whead)) ||
		(whead.head != C4_DWORD('R','I','F','F'))) {
		file_close(src);
		return(FAILURE);
	}
	if ((file_read(src, &whead, 4) != 4) ||
		(whead.head != C4_DWORD('W','A','V','E'))) {
		file_close(src);
		return(FAILURE);
	}
	if ((file_read(src, &whead, sizeof(whead)) != sizeof(whead)) ||
		(whead.head != C4_DWORD('f','m','t',' ')) ||
		(file_read(src, &winfo, sizeof(winfo)) != sizeof(winfo))) {
		file_close(src);
		return(FAILURE);
	}
	seekp = whead.size + 20;
	file_seek(src, seekp, FSEEK_SET);
	if (file_read(src, &whead, sizeof(whead)) != sizeof(whead)) {
		file_close(src);
		return(FAILURE);
	}
	if (whead.head == C4_DWORD('f','a','c','t')) {
		seekp += whead.size + 8;
		file_seek(src, seekp, FSEEK_SET);
		if (file_read(src, &whead, sizeof(whead)) != sizeof(whead)) {
			file_close(src);
			return(FAILURE);
		}
	}
	if (whead.head != C4_DWORD('d','a','t','a')) {
		file_close(src);
		return(FAILURE);
	}

	// tH[}bg`FbN
	if ((winfo.channel != 1) ||
		(winfo.rate != 44100) ||
		(winfo.bit != 16)) {
		file_close(src);
		return(FAILURE);
	}

	step = 44100 / samplerate;
	item->rhythm_samples = whead.size / (winfo.bit / 8) / winfo.channel / step;

	// waveǂݍ݃A[P[g
	if ((hwave = GlobalAlloc(GPTR, whead.size)) == NULL) {
		file_close(src);
		return(FAILURE);
	}
	if ((wave = (BYTE *)GlobalLock(hwave)) == NULL) {
		GlobalFree(hwave);
		file_close(src);
		return(FAILURE);
	}

	// wavef[^ǂݍ
	if (file_lread(src, wave, whead.size) != whead.size) {
		GlobalUnlock(hwave);
		GlobalFree(hwave);
		file_close(src);
	}
	file_close(src);

	// f[^̃A[P[g
	if ((item->handle = GlobalAlloc(GPTR, item->rhythm_samples * 2)) == NULL) {
		GlobalUnlock(hwave);
		GlobalFree(hwave);
		return(FAILURE);
	}
	if ((item->rhythm_data = (BYTE *)GlobalLock(item->handle)) == NULL) {
		GlobalFree(item->handle);
		GlobalUnlock(hwave);
		GlobalFree(hwave);
		item->handle = NULL;
		return(FAILURE);
	}

	// f[^̃Rs[
	for (i=0; i<item->rhythm_samples; i++) {
		*(short *)(&item->rhythm_data[i*2]) = *(short *)(&wave[i*step*2]);
	}

	// 
	GlobalUnlock(hwave);
	GlobalFree(hwave);
	return(SUCCESS);
}


static char *rhythmfile[RHYTHM_MAX] = {
						"2608_bd.wav",
						"2608_sd.wav",
						"2608_top.wav",
						"2608_hh.wav",
						"2608_tom.wav",
						"2608_rim.wav"};


void rhythm_init(void) {

	int		i;

	ZeroMemory(&rhythm, sizeof(rhythm));

	for (i=0; i<RHYTHM_MAX; i++) {
		waveread(&rhythm.r[i], rhythmfile[i], opna_rate);
	}
}


void rhythm_term(void) {

	int		i;

	for (i=0; i<RHYTHM_MAX; i++) {
		if (rhythm.r[i].handle) {
			GlobalUnlock(rhythm.r[i].handle);
			GlobalFree(rhythm.r[i].handle);
			rhythm.r[i].handle = NULL;
		}
	}
}


void rhythm_reset(void) {

	rhythm.bitmap = 0;
}


static void rhythm_play(int num, int event) {

	if ((num < RHYTHM_MAX) && (rhythm.r[num].handle)) {
		if (!event) {
			rhythm.bitmap |= (1 << num);
			rhythm.r[num].rhythm_ptr = rhythm.r[num].rhythm_data;
			rhythm.r[num].rhythm_remain = rhythm.r[num].rhythm_samples;
		}
		else {
			rhythm.bitmap &= ~(1 << num);
		}
	}
}

void rhythm_setregister(BYTE reg, BYTE value) {

	BYTE	bit;
	int		i;

	if (reg == 0x10) {
		DSound_Send0();
		for (i=0, bit=1; i<6; i++, bit<<=1) {
			if (value & bit) {
				rhythm_play(i, value & 0x80);
			}
		}
	}
	else if ((reg >= 0x18) && (reg < 0x1e)) {
		rhythm.r[reg-0x18].rhythm_volume = (~value) & 0x1f;
		rhythm.r[reg-0x18].rhythm_lr = value & 0xc0;
	}
}

LABEL void __fastcall rhythm_update(void *buf, int buflength) {

		__asm {
				cmp		edx, 0
				je		rhythm_end
				cmp		rhythm.bitmap, 0
				je		rhythm_end
				pushad
				xor		bx, bx
				mov		edi, offset (rhythm.r[0])
rhythm_lp:		bt		rhythm.bitmap, bx
				jnc		rhythm_next
				mov		ebp, [edi]RHYTHMFILE_TABLE.rhythm_remain
				cmp		edx, ebp
				jae		rhythm_remainad
				mov		ebp, edx
rhythm_remainad:sub		[edi]RHYTHMFILE_TABLE.rhythm_remain, ebp
				jg		rhythm_outst
				btr		rhythm.bitmap, bx
rhythm_outst:	mov		esi, [edi]RHYTHMFILE_TABLE.rhythm_ptr
				push	ecx
				push	edx
rhythm_outlp:	movsx	eax, word ptr [esi]
				imul	[edi]RHYTHMFILE_TABLE.rhythm_volume
				sar		eax, 6
				test	[edi]RHYTHMFILE_TABLE.rhythm_lr, 80h
				je		rhythm_outr
				add		[ecx], eax
rhythm_outr:	test	[edi]RHYTHMFILE_TABLE.rhythm_lr, 40h
				je		rhythm_outlped
				add		[ecx+4], eax
rhythm_outlped:	add		esi, 2
				add		ecx, 8
				dec		ebp
				jne		rhythm_outlp
				pop		edx
				pop		ecx
				mov		[edi]RHYTHMFILE_TABLE.rhythm_ptr, esi
rhythm_next:	add		edi, (type RHYTHMFILE_TABLE)
				inc		bx
				cmp		bx, RHYTHM_MAX
				jc		rhythm_lp
				popad
rhythm_end:		ret
		}
}
